#This file is for all the functions used in the project.
#it needs to be run before the main files and the data file

AllFunc = T
if(AllFunc){
  library(TMB)
  library(skellam)
  greys <- grep("gr[ea]y", colours(), value = TRUE)
  #a function that takes two dates and give the absoulut difference in days
  differenceDates <- function(date1,date2){
    x = as.numeric(as.POSIXct(date1, format="%d/%m/%Y"))
    y = as.numeric(as.POSIXct(date2, format="%d/%m/%Y"))
    z = abs(x-y)
    return(floor(z/(24*3600)))
  }
  
  #a function that takes the data set and the different teams as input and returns a matrix where each colums is what match it is for the home or away team
  matchNumber <- function(pl=pl,teams){
    m = matrix(0,nrow=length(pl$HomeTeam),ncol=2)
    team = cbind(pl$HomeTeam,pl$AwayTeam)
    #teams = sort(levels(factor(team[,1])))
    count = rep(0,nlevels(factor(teams)))
    
    for(i in 1:length(team[,1])){
      xh = which(team[i,1]==teams)
      xa = which(team[i,2]==teams)
      m[i,1] = count[xh]
      m[i,2] = count[xa]
      count[xh] = count[xh]+1
      count[xa] = count[xa] +1
    }
    return(m)
  }
  #a home away function for one season at a time only
  #row tells what team is playing at home and what team is playing away at match i
  HomeAway <- function(pl){
    m = length(pl$HomeTeam)
    home=rep(0,m)
    away = rep(0,m)
    teams = sort(levels(factor(pl$HomeTeam)))
    for(i in 1:m){
      home[i] = which(pl$HomeTeam[i]==teams)
      away[i] = which(pl$AwayTeam[i]==teams)
    }
    home=factor(home-1)
    away = factor(away-1)
    return(cbind(home,away))
  }
  #a function that simulates results based on the estimated probabilities
  #it is used to test the models
  randomResults <-function(probs){
    pr = c("H","D","A")
    n = length(probs[,1])
    results = rep("H",n)
    for(i in 1:n){
      probs1 = probs[i,]
      x = rmultinom(1,1,probs1)[,1]
      m = which(x==1)
      results[i]=pr[m]
    }
    return(results)
  } 
  #this function is meant to help the function below
  poisToProbs <- function(home,away,rep){
    home = as.numeric(home)
    away = as.numeric(away)
    H = rep(0,length(home))
    A = rep(0,length(home))
    a = rep$par.random[1:20]
    d = rep$par.random[21:40]
    c = rep$par.fixed[3]
    ch = rep$par.fixed[4]
    for(i in 1:length(home)){
      
      H[i] = (c+ch+a[home[i]]-d[away[i]])
      A[i] = (c-d[home[i]]+a[away[i]])
      
    }
    
    #now I have all the expected number of goals
    return(cbind(1-pskellam(0,exp(H),exp(A)),dskellam(0,exp(H),exp(A)),pskellam(-1,exp(H),exp(A))))
  }
  #This function gives the estimated probabilities for the next M rounds after round 19 of the seaons of choice.
  #using the Poisson GLMM 
  timePoisBIToProbs <- function(PL,M){
    #dyn.load(dynlib("binormalintercept"))
    random=c("a","d")
    parameters = list(log_sd = c(0.1,0.1),
                      a=rep(0,20),d=rep(0,20),c=0.5,ch=0.5)
    #always testing for the last half of the season
    probs = cbind(rep(0,M*10),rep(0,M*10),rep(0,M*10))
    x = PL$FTHG
    y = PL$FTAG
    HA = HomeAway(PL)
    home = as.factor(HA[,1]-1)
    away = as.factor(HA[,2]-1)
    for(i in 1:M){
      data = list(x = x[1:(190+(i-1)*10)],y = y[1:(190+(i-1)*10)],home=home[1:(190+(i-1)*10)],away=away[1:(190+(i-1)*10)],N=nlevels(home))
      obj <- MakeADFun(data=data,parameters=parameters,random = random,DLL = "binormalintercept", silent = TRUE)
      res <- optim(par = obj$par, fn = obj$fn, gr = obj$gr, method = "BFGS")
      rep = sdreport(obj)
      probs[((i-1)*10+1):(i*10),] = poisToProbs(home[(1+190+(i-1)*10):(190+(i)*10)],away[(1+190+(i-1)*10):(190+(i)*10)],rep)
    }
    return(probs)
  }
  #this function gives the betting results based on the odds, the results and the estimated probabilities.
  #it used the fixed bet strategy, meaning that one wagers the same at every wager with expected profit.
  fixedBet <- function(probs,odds,C,results){
    ER = probs*odds
    n = length(probs[,1])/10
    coin = rep(0,n+1)
    for(i in 1:n){
      coin1 = -C
      prob1 = probs[(10*(i-1)+1):(10*i),]
      ER1 = ER[(10*(i-1)+1):(10*i),]
      res1 = results[(10*(i-1)+1):(10*i)]
      m = length(which(ER1>1))
      odds1 = odds[(10*(i-1)+1):(10*i),]
      for(j in 1:10){
        if(ER1[j,1]>1&res1[j]=="H"){
          coin1 = coin1 + odds1[j,1]*C/m
          #print(c(ER1[j,1],res1[j],1))
        }
        
        if(ER1[j,2]>1&res1[j]=="D"){
          coin1 = coin1 + odds1[j,2]*C/m
          #print(c(ER1[j,2],res1[j],2))
        }
        if(ER1[j,3]>1&res1[j]=="A"){
          coin1 = coin1 + odds1[j,3]*C/m
          #print(c(ER1[j,3],res1[j],3))
        }
      }
      coin[i+1]=coin[i]+coin1
    }
    return(coin)
  }
  #the same as the previous, but this one uses fixed return.
  FixedReturn <-function(probs,odds,C,results){
    ER = probs*odds
    n = length(probs[,1])/10
    coin = rep(0,n+1)
    for(i in 1:n){
      coin1 = -C
      prob1 = probs[(10*(i-1)+1):(10*i),]
      ER1 = ER[(10*(i-1)+1):(10*i),]
      res1 = results[(10*(i-1)+1):(10*i)]
      m = length(which(ER1>1))
      odds1 = odds[(10*(i-1)+1):(10*i),]
      q = 0
      wins = 0
      for(j in 1:10){
        if(ER1[j,1]>1){
          q = q + 1/odds1[j,1]
          if(res1[j]=="H"){
            wins= wins +1
          }
          
        }
        
        if(ER1[j,2]>1){
          q = q + 1/odds1[j,2]
          if(res1[j]=="D"){
            wins= wins +1
          }
          
        }
        if(ER1[j,3]>1){
          q = q + 1/odds1[j,3]
          if(res1[j]=="A"){
            wins= wins +1
          }
          
        }
      }
      k = C/q
      coin[i+1] = coin[i]+coin1 + k*wins
    }
    return(coin)
  }
  #kelly criterion, beta is the factor to multiply the fraction chosen by Kelly criterion. It should be less than 1 
  testKelly <-function(probs,odds,C,results,beta){
    n=length(results)
    ER = probs*odds
    res = c("H","D","A")
    coins = rep(0,n+1)
    coins[1] = C
    for(i in 1:n){
      dC = 0
      m = which.max(ER[i,])
      if(ER[i,m]>1){
        frac = probs[i,m]-(1-probs[i,m])/(odds[i,m]-1)
        if(results[i]==res[m]){
          dC = (odds[i,m]-1)*coins[i]*frac*beta
        }
        else{
          dC = -frac*coins[i]*beta
        }
      }
      coins[i+1] = coins[i]+dC
    }
    return(coins)
  }
  # the rue and salvesen betting strategy
  RueVariance <-function(probs,odds,C,results){
    ER = probs*odds
    n = length(probs[,1])/10
    coin = rep(0,n+1)
    res = c("H","D","A")
    for(i in 1:n){
      coin1 = -C
      prob1 = probs[(10*(i-1)+1):(10*i),]
      ER1 = ER[(10*(i-1)+1):(10*i),]
      res1 = results[(10*(i-1)+1):(10*i)]
      beta = rep(0,10)
      odds1 = odds[(10*(i-1)+1):(10*i),]
      wo = rep(0,10)
      for(j in 1:10){
        m = which.max(ER1[j,])
        if(ER1[j,m]>1){
          beta[j] = (ER1[j,m]-1)/(2*odds1[j,m]^2*prob1[j,m]*(1-prob1[j,m]))
          if(res1[j]==res[m]){
            wo[j]=odds1[j,m]*beta[j]
          }
        }
        
      }
      k = C/sum(beta)
      coin1 = coin1 + k*sum(wo)
      coin[i+1] = coin[i]+coin1 
    }
    return(coin)
  }
  #gives a matrix where each row is the numer of days between each match for one team
  makeDTfull <- function(EPL,teams,n){
    DT = matrix(10^(-5),ncol = n-1,nrow = length(teams))
    for(i in 1:length(teams)){
      #print(i)
      m = sum(EPL$HomeTeam==teams[i])+sum(EPL$AwayTeam==teams[i])
      dato = EPL$Date[which(EPL$HomeTeam==teams[i]|EPL$AwayTeam==teams[i])]
      if(m>1){
        for(t in 1:(m-1)){
          #print(t)
          DT[i,t] = differenceDates(dato[t],dato[t+1])
        }
      }
      
    }
    return(DT)
  }
  #returns a matrix where each colum says what team is playing the match.
  #this is so that it can be used in c++ code.
  HomeAwayFull <- function(pl,teams){
    m = length(pl$HomeTeam)
    home=rep(-1,m)
    away = rep(-1,m)
    for(i in 1:m){
      home[i] = which(pl$HomeTeam[i]==teams)
      away[i] = which(pl$AwayTeam[i]==teams)
    }
    return(cbind(home,away))
  }
  #this function returns probabilities for the results using the poisson model with strength modeld as a brownina motion. 
  #the brwnian motion uses a fixed tau.
  fullToProbsTau <- function(PL,start,slutt,n1,M1,tau){
    #start er frste runde som skal ha sannsynligheter 
    #slutt er site runde som skal ha sannsynligheter 
    #start=20,slutt=38. gir sannsynligheter for siste halvdel av sesongen
    #n1 er antall kamper som ikke skal vre med for prediksjonen. n1=0 gir hele datasettet
    #PL[(n1+1):M1,] er datasettet som skal brukes til frste prediksjon
    #for  bruke en sesong og prediktere andre halvdel s skal n1=0 og m1=190
    
    probs = matrix(0,ncol = 3,nrow = (slutt-start+1)*10)
    #i=8
    startTime = Sys.time()
    temp = Sys.time()
    k=slutt-start+1
    for(i in 1:k){
      print(i)
      print( Sys.time()-temp)
      temp = Sys.time()
      EPL = PL[(n1+1):(M1+10*(i-1)),]
      EPL2 = PL[(M1+10*(i-1)+1):(M1+10*(i)),]
      teams = levels(as.factor(PL$HomeTeam[1:(M1+10*(k))]))
      nt = length(teams)#the number of different teams in the dataset
      n = length(c(EPL$Date[which(EPL$HomeTeam==teams[1])],EPL$Date[which(EPL$AwayTeam==teams[1])]))#the total number of matches played by each team that has been in the top every season
      for(j in 2:nt){
        n = max(n, length(c(EPL$Date[which(EPL$HomeTeam==teams[j])],EPL$Date[which(EPL$AwayTeam==teams[j])])))
      }
      HA = HomeAwayFull(EPL,teams)
      home = as.factor(HA[,1]-1)
      away = as.factor(HA[,2]-1)
      x = EPL$FTHG
      y = EPL$FTAG
      
      data = list(x=x,y=y,home=home,away=away,nt=nt)
      parameters = list(a0=rep(0,nt),d0=rep(0,nt),log_sd=c(-1,-1),c=0.5,ch=0.5)
      
      random = c("a0","d0")
      
      obj <- MakeADFun(data=data,parameters=parameters,random = random,DLL = "FullModelConst", silent = TRUE)
      res <- optim(par = obj$par, fn = obj$fn, gr = obj$gr, method = "BFGS")
      
      repC = sdreport(obj)
      #print(repC)
      #sn = as.factor(rep(0:1,each=4088))
      a0 = repC$par.random[1:nt]
      d0 = repC$par.random[(nt+1):(2*nt)]
      log_sd = repC$par.fixed[1:2]
      c = repC$par.fixed[3]
      ch = repC$par.fixed[4]
      a = matrix(0,nrow = nt,ncol=n-1)
      d = matrix(0,nrow = nt,ncol=n-1)
      for(j in 1:nt){
        a[j,] = rep(a0[j],n-1)
        d[j,] = rep(d0[j],n-1)
      }
      
      HA = HomeAwayFull(EPL,teams)
      m = matchNumber(EPL,teams)
      data = list(x = EPL$FTHG,y=EPL$FTAG,home = as.factor(HA[,1]-1),away=as.factor(HA[,2]-1),DT = sqrt(makeDTfull(EPL,teams,n)),
                  mh = as.factor(m[,1]),ma = as.factor(m[,2]),N=n-1,nt=nt,tau=tau)
      
      parameters = list(a0=a0,d0= d0,a = a,d = d,log_sd=log_sd,
                        c=c,ch=ch,2)
      
      random = c("a0","d0","a","d")
      
      obj <- MakeADFun(data=data,parameters=parameters,random = random,DLL = "FullModelTau", silent = TRUE)
      res <- optim(par = obj$par, fn = obj$fn, gr = obj$gr, method = "BFGS")
      rep = sdreport(obj)
      
      a = matrix(rep$par.random[(2*nt+1):(2*nt+nt*(n-1))],ncol=n-1,nrow = nt)
      a0 = rep$par.random[1:nt]
      d0 = rep$par.random[(nt+1):(2*nt)]
      d = matrix(rep$par.random[(2*nt+nt*(n-1)+1):(2*nt+2*nt*(n-1))],ncol=n-1,nrow = nt)
      
      m1 = matchNumber(rbind(EPL,EPL2),teams)
      mh1 = pmin(m1[(M1+10*(i-1)+1):(M1+10*(i)),1]-1,n-1)
      ma1 = pmin(m1[(M1+10*(i-1)+1):(M1+10*(i)),2]-1,n-1)
      HA1 = HomeAwayFull(EPL2,teams)
      home1 = HA1[,1]
      away1 = HA1[,2]
      for(j in 1:10){
        lamh = (a[home1[j],mh1[j]]-d[away1[j],ma1[j]]+c+ch)
        lama = (a[away1[j],ma1[j]]-d[home1[j],mh1[j]]+c)
        
        probs[(i-1)*10+j,]=cbind(1-pskellam(0,exp(lamh),exp(lama)),dskellam(0,exp(lamh),exp(lama)),pskellam(-1,exp(lamh),exp(lama)))
      }
      
    }
    #return(cbind(a0,d0,a,d))
    print(Sys.time()-startTime)
    return(probs)
  }
  #this function returns probabilities for the results using the poisson model with strength modeld as a brownina motion. 
  #the brwnian motion uses a tau estimated in the likelihood.
  fullToProbsTauMLE <- function(PL,start,slutt,n1,M1){
    #start er frste runde som skal ha sannsynligheter 
    #slutt er site runde som skal ha sannsynligheter 
    #start=20,slutt=38. gir sannsynligheter for siste halvdel av sesongen
    #n1 er antall kamper som ikke skal vre med for prediksjonen. n1=0 gir hele datasettet
    #PL[(n1+1):M1,] er datasettet som skal brukes til frste prediksjon
    #for  bruke en sesong og prediktere andre halvdel s skal n1=0 og m1=190
    
    probs = matrix(0,ncol = 3,nrow = (slutt-start+1)*10)
    #i=1
    startTime = Sys.time()
    temp = Sys.time()
    k=slutt-start+1
    for(i in 1:k){
      print(i)
      print( Sys.time()-temp)
      temp = Sys.time()
      EPL = PL[(n1+1):(M1+10*(i-1)),]
      EPL2 = PL[(M1+10*(i-1)+1):(M1+10*(i)),]
      teams = levels(as.factor(PL$HomeTeam[1:(M1+10*(k))]))
      nt = length(teams)#the number of different teams in the dataset
      n = length(c(EPL$Date[which(EPL$HomeTeam==teams[1])],EPL$Date[which(EPL$AwayTeam==teams[1])]))#the total number of matches played by each team that has been in the top every season
      for(j in 2:nt){
        n = max(n, length(c(EPL$Date[which(EPL$HomeTeam==teams[j])],EPL$Date[which(EPL$AwayTeam==teams[j])])))
      }
      HA = HomeAwayFull(EPL,teams)
      home = as.factor(HA[,1]-1)
      away = as.factor(HA[,2]-1)
      x = EPL$FTHG
      y = EPL$FTAG
      
      data = list(x=x,y=y,home=home,away=away,nt=nt)
      parameters = list(a0=rep(0,nt),d0=rep(0,nt),log_sd=c(-1,-1),c=0.5,ch=0.5)
      
      random = c("a0","d0")
      
      obj <- MakeADFun(data=data,parameters=parameters,random = random,DLL = "FullModelConst", silent = TRUE)
      res <- optim(par = obj$par, fn = obj$fn, gr = obj$gr, method = "BFGS")
      
      repC = sdreport(obj)
      #print(repC)
      #sn = as.factor(rep(0:1,each=4088))
      a0 = repC$par.random[1:nt]
      d0 = repC$par.random[(nt+1):(2*nt)]
      log_sd = repC$par.fixed[1:2]
      c = repC$par.fixed[3]
      ch = repC$par.fixed[4]
      a = matrix(0,nrow = nt,ncol=n-1)
      d = matrix(0,nrow = nt,ncol=n-1)
      for(j in 1:nt){
        a[j,] = rep(a0[j],n-1)
        d[j,] = rep(d0[j],n-1)
      }
      
      HA = HomeAwayFull(EPL,teams)
      m = matchNumber(EPL,teams)
      data = list(x = EPL$FTHG,y=EPL$FTAG,home = as.factor(HA[,1]-1),away=as.factor(HA[,2]-1),DT = sqrt(makeDTfull(EPL,teams,n)),
                  mh = as.factor(m[,1]),ma = as.factor(m[,2]),N=n-1,nt=nt)
      
      parameters = list(a0=a0,d0= d0,a = a,d = d,log_sd=log_sd,
                        c=c,ch=ch,tau=55)
      
      random = c("a0","d0","a","d")
      
      obj <- MakeADFun(data=data,parameters=parameters,random = random,DLL = "FullModel", silent = TRUE)
      res <- optim(par = obj$par, fn = obj$fn, gr = obj$gr, method = "BFGS")
      rep = sdreport(obj)
      
      a = matrix(rep$par.random[(2*nt+1):(2*nt+nt*(n-1))],ncol=n-1,nrow = nt)
      a0 = rep$par.random[1:nt]
      d0 = rep$par.random[(nt+1):(2*nt)]
      d = matrix(rep$par.random[(2*nt+nt*(n-1)+1):(2*nt+2*nt*(n-1))],ncol=n-1,nrow = nt)
      
      m1 = matchNumber(rbind(EPL,EPL2),teams)
      mh1 = pmin(m1[(M1+10*(i-1)+1):(M1+10*(i)),1]-1,n-1)
      ma1 = pmin(m1[(M1+10*(i-1)+1):(M1+10*(i)),2]-1,n-1)
      HA1 = HomeAwayFull(EPL2,teams)
      home1 = HA1[,1]
      away1 = HA1[,2]
      for(j in 1:10){
        lamh = (a[home1[j],mh1[j]]-d[away1[j],ma1[j]]+c+ch)
        lama = (a[away1[j],ma1[j]]-d[home1[j],mh1[j]]+c)
        
        probs[(i-1)*10+j,]=cbind(1-pskellam(0,exp(lamh),exp(lama)),dskellam(0,exp(lamh),exp(lama)),pskellam(-1,exp(lamh),exp(lama)))
      }
      
    }
    #return(cbind(a0,d0,a,d))
    print(Sys.time()-startTime)
    return(probs)
  }
  #this function returns probabilities for the results using the generalized bradley-terry  model with strength modeld as a brownian motion. 
  #the brownian motion uses a tau estimated in the likelihood.
  fullMultinomToProbs <- function(PL,start,slutt,n1,M1){
    #start er frste runde som skal ha sannsynligheter 
    #slutt er site runde som skal ha sannsynligheter 
    #start=20,slutt=38. gir sannsynligheter for siste halvdel av sesongen
    #n1 er antall kamper som ikke skal vre med for prediksjonen. n1=0 gir hele datasettet
    #PL[(n1+1):M1,] er datasettet som skal brukes til frste prediksjon
    #for  bruke en sesong og prediktere andre halvdel s skal n1=0 og m1=190
    
    probs = matrix(0,ncol = 3,nrow = (slutt-start+1)*10)
    #i=8
    startTime = Sys.time()
    temp = Sys.time()
    k=slutt-start+1
    for(i in 1:k){
      print(i)
      print( Sys.time()-temp)
      temp = Sys.time()
      EPL = PL[(n1+1):(M1+10*(i-1)),]
      EPL2 = PL[(M1+10*(i-1)+1):(M1+10*(i)),]
      teams = levels(as.factor(PL$HomeTeam[1:(M1+10*(k))]))
      nt = length(teams)#the number of different teams in the dataset
      n = length(c(EPL$Date[which(EPL$HomeTeam==teams[1])],EPL$Date[which(EPL$AwayTeam==teams[1])]))#the total number of matches played by each team that has been in the top every season
      for(j in 2:nt){
        n = max(n, length(c(EPL$Date[which(EPL$HomeTeam==teams[j])],EPL$Date[which(EPL$AwayTeam==teams[j])])))
      }
      HA = HomeAwayFull(EPL,teams)
      home = as.factor(HA[,1]-1)
      away = as.factor(HA[,2]-1)
      X = cbind(as.numeric(EPL$FTR=="H"),as.numeric(EPL$FTR=="D"),as.numeric(EPL$FTR=="A"))
      
      data = list(X=X,home=home,away=away)
      parameters = list(lam = rep(0,nt),log_sigma=-1,theta = 0.6,h=0.4)
      
      random = c("lam")
      
      obj <- MakeADFun(data=data,parameters=parameters,random = random,DLL = "multinomRI", silent = TRUE)
      res <- optim(par = obj$par, fn = obj$fn, gr = obj$gr, method = "BFGS")
      
      repC = sdreport(obj)
      #print(repC)
      #sn = as.factor(rep(0:1,each=4088))
      lam1 = repC$par.random[1:nt]
      log_sigma = repC$par.fixed[3]
      theta = repC$par.fixed[1]
      h = repC$par.fixed[2]
      lam = matrix(0,nrow = nt,ncol=n-1)
      for(j in 1:nt){
        lam[j,] = rep(lam1[j],n-1)
      }
      
      HA = HomeAwayFull(EPL,teams)
      m = matchNumber(EPL,teams)
      data = list(X=X,home = as.factor(HA[,1]-1),away=as.factor(HA[,2]-1),DT = sqrt(makeDTfull(EPL,teams,n)),
                  mh = as.factor(m[,1]),ma = as.factor(m[,2]),N=n-1,nt=nt)
      
      parameters = list(theta=theta,h=h,lam1=lam1,lam=lam,log_sigma=log_sigma,tau=45)
      
      random = c("lam1","lam")
      
      obj <- MakeADFun(data=data,parameters=parameters,random = random,DLL = "multinomRW", silent = TRUE)
      res <- optim(par = obj$par, fn = obj$fn, gr = obj$gr, method = "BFGS")
      rep = sdreport(obj)
      
      lam = matrix(rep$par.random[(nt+1):(nt+nt*(n-1))],ncol=n-1,nrow = nt)
      lam1 = rep$par.random[1:nt]
      
      m1 = matchNumber(rbind(EPL,EPL2),teams)
      mh1 = pmin(m1[(M1+10*(i-1)+1):(M1+10*(i)),1]-1,n-1)
      ma1 = pmin(m1[(M1+10*(i-1)+1):(M1+10*(i)),2]-1,n-1)
      HA1 = HomeAwayFull(EPL2,teams)
      home1 = HA1[,1]
      away1 = HA1[,2]
      for(j in 1:10){
        ph = plogis(lam[home1[j],mh1[1]]-lam[away1[j],ma1[1]]+h-theta)
        pd = plogis(lam[home1[j],mh1[1]]-lam[away1[j],ma1[1]]+h+theta) - plogis(lam[home1[j],mh1[1]]-lam[away1[j],ma1[1]]+h-theta)
        pa = 1 - plogis(lam[home1[j],mh1[1]]-lam[away1[j],ma1[1]]+h+theta)
        
        probs[(i-1)*10+j,]=c(ph,pd,pa)
      }
      
    }
    #return(cbind(a0,d0,a,d))
    print(Sys.time()-startTime)
    return(probs)
  }
  #a function that  gives the pmf value for the generlized poisson 
  dGP <- function(x,lam,delta){
    return(dpois(x,lam)*((lam+delta*x)/lam)^(x-1)*exp(-delta*x))
  }
  #a function that gives the cumulative value for a generalized poisson distributed variabel
  pGP <- function(x,lam,delta){
    s = 0
    for(i in 0:x){
      s = s + dGP(i,lam,delta)
    }
    return(s)
  }
  #using the estimated values from TMB it predicts the future probabilities.
  GPtoProbs <- function(home,away,c,ch,atk,def,delta){
    home = as.numeric(home)
    away = as.numeric(away)
    H = rep(0,length(home))
    A = rep(0,length(home))
    for(i in 1:length(home)){
      H[i] = exp(c+ch+atk[home[i]]-def[away[i]])
      A[i] = exp(c-def[home[i]]+atk[away[i]])
    }
    probs = matrix(0,ncol=3,nrow=length(home))
    for(i in 1:length(home)){
      ph = 0
      pd = 0
      pa = 0
      for(t in 0:10){
        ph = ph + (1-pGP(t,H[i],delta))*dGP(t,A[i],delta)
        pd = pd + dGP(t,H[i],delta)*dGP(t,A[i],delta)
        pa = pa + (1-pGP(t,A[i],delta))*dGP(t,H[i],delta)
        #ph = ph + (1-ppois(t,H[i]))*dpois(t,A[i])
        #pd = pd + dpois(t,H[i])*dpois(t,A[i])
        #pa = pa + (1-ppois(t,A[i]))*dpois(t,H[i])
      }
      probs[i,]=c(ph,pd,pa)
    }
    return(probs)
  }
  #predicts the probabilities for the next M rounds after round 19 in the season. 
  timeGPtoProbs<- function(PL,M){
    #dyn.load(dynlib("GP"))
    HA = HomeAway(PL)
    home = as.factor(HA[,1]-1)
    away = as.factor(HA[,2]-1)
    x = PL$FTHG
    y = PL$FTAG
    random = c("a","d")
    parameters = list(log_sd=c(-1,-1),a=rep(0,20),d=rep(0,20),c=0.5,ch=0.5,delta=0)
    #always testing for the last half of the season
    probs = cbind(rep(0,10*M),rep(0,10*M),rep(0,10*M))
    for(i in 1:M){
      data = list(x = x[1:(190+(i-1)*10)],y = y[1:(190+(i-1)*10)],home=home[1:(190+(i-1)*10)],away=away[1:(190+(i-1)*10)],x=x[1:(190+(i-1)*10)],
                  y=y[1:(190+(i-1)*10)],N=nlevels(home))
      obj <- MakeADFun(data=data,parameters=parameters,random = random,DLL = "GP", silent = TRUE)
      res <- optim(par = obj$par, fn = obj$fn, gr = obj$gr, method = "BFGS")
      rep = sdreport(obj)
      probs[((i-1)*10+1):(i*10),] = GPtoProbs(home[(1+190+(i-1)*10):(190+(i)*10)],away[(1+190+(i-1)*10):(190+(i)*10)],rep$par.fixed[3],rep$par.fixed[4],rep$par.random[1:20],rep$par.random[21:40],rep$par.fixed[5])
    }
    return(probs)
  }
  #estimating the probabilities for the next M rounds after round 19 using the data intensive model.
  HLCtoProbs <- function(PL,M){
    #dyn.load(dynlib("HLC2"))
    HA = HomeAway(PL)
    home = as.factor(HA[,1]-1)
    away = as.factor(HA[,2]-1)
    Sh = pmax(PL$HS,PL$HST)
    Sa = pmax(PL$AS,PL$AST)
    Th = pmax(PL$FTHG,PL$HST)
    Ta =pmax(PL$FTAG,PL$AST)
    x = PL$FTHG
    y = PL$FTAG
    probs = matrix(0,nrow = 10*M,ncol = 3)
    for(i in 1:M){
      k = 190+(i-1)*10
      
      data = list(x=x[1:k],y=y[1:k],home=home[1:k],away=away[1:k],Sh=Sh[1:k],Sa=Sa[1:k],Th=Th[1:k],Ta=Ta[1:k])
      parameters = list(log_lambda=rep(3,20),a=rep(0,20),b=rep(0,20),g=rep(0,20),log_h=0)
      
      obj <- MakeADFun(data=data,parameters=parameters,DLL = "HLC2", silent = TRUE)
      res <- optim(par = obj$par, fn = obj$fn, gr = obj$gr, method = "BFGS")
      rep = sdreport(obj)
      a = pnorm(rep$par.fixed[1:20])
      b = pnorm(rep$par.fixed[21:40])
      g = pnorm(rep$par.fixed[41:60])
      lambda = exp(rep$par.fixed[61:80])
      h = exp(rep$par.fixed[81])
      home1 = as.numeric(home[(k+1):(k+10)])
      away1 = as.numeric(away[(k+1):(k+10)])
      for(j in 1:10){
        lamh = h*lambda[home1[j]]*a[home1[j]]*b[away1[j]]*g[away1[j]]
        lama = lambda[away1[j]]*a[away1[j]]*b[home1[j]]*g[home1[j]]
        probs[(i-1)*10+j,]=c(1-pskellam(0,lamh,lama),dskellam(0,lamh,lama),pskellam(-1,lamh,lama))
      }
    }
    return(probs)
  }
  
  #the multinom constant function
  multinomToProbs <- function(PL,M){
    #dyn.load(dynlib("multinomRI"))
    parameters = list(theta=0.1,h=0,lam=rep(0,20),log_sigma=-1)
    random = c("lam")
    #M is the number of rounds to be estimated, it always starts at round 20. 
    probs = matrix(0,nrow=10*M,ncol = 3)
    HA = HomeAway(PL)
    home = as.factor(HA[,1]-1)
    away = as.factor(HA[,2]-1)
    X = cbind(as.numeric(PL$FTR=="H"),as.numeric(PL$FTR=="D"),as.numeric(PL$FTR=="A"))
    
    for(i in 1:M){
      data = list(X=X[1:(190+(i-1)*10),],home=home[1:(190+(i-1)*10)],away=away[1:(190+(i-1)*10)])
      obj <- MakeADFun(data=data,parameters=parameters,random = random,DLL = "multinomRI", silent = TRUE)
      res <- optim(par = obj$par, fn = obj$fn, gr = obj$gr, method = "BFGS")
      rep = sdreport(obj)
      lam = rep$par.random[1:20]
      theta = rep$par.fixed[1]
      h = rep$par.fixed[2]
      home1 = HA[(190+(i-1)*10+1):(190+(i)*10),1]
      away1 = HA[(190+(i-1)*10+1):(190+(i)*10),2]
      for(j in 1:10){
        ph = plogis(h-theta+lam[home1[j]]-lam[away1[j]])
        pd = plogis(h+theta+lam[home1[j]]-lam[away1[j]])-plogis(h-theta+lam[home1[j]]-lam[away1[j]])
        pa = 1-plogis(h+theta+lam[home1[j]]-lam[away1[j]])
        probs[(i-1)*10+j,]=c(ph,pd,pa)
      }
    }
    return(probs)
  }
  #this function plots the odds, in a linear fashion
  plotOdds <- function(odds,probs,results,season="",model="",HUB=F){
    color = c("green","red")
    HUB = c("H","D","A")
    ER = odds*probs
    res = matrix(0,ncol = 3,nrow = length(results))
    for(i in 1:length(probs[,1])){
      for(j in 1:length(probs[1,])){
        if(ER[i,j]>1){
          res[i,j]=2*(results[i]==HUB[j])-1
        }
      }
    }
    if(HUB==T){
      plot(odds[which(res[,1]==1),1],probs[which(res[,1]==1),1],col="green",xlim = c(0,25),pch=16,ylim = c(0,1),
           main=paste("Season=",season," and model=",model),xlab = "Odds",ylab = "Estimated Probability",cex=0.7)
      points(odds[which(res[,1]==-1),1],probs[which(res[,1]==-1),1],col="red",pch=16,cex=0.7)
      points(odds[which(res[,2]==-1),2],probs[which(res[,2]==-1),2],col="red",pch=17,cex=0.7)
      points(odds[which(res[,3]==-1),3],probs[which(res[,3]==-1),3],col="red",pch=18,cex=0.7)
      points(odds[which(res[,2]==1),2],probs[which(res[,2]==1),2],col="green",pch=17,cex=0.7)
      points(odds[which(res[,3]==1),3],probs[which(res[,3]==1),3],col="green",pch=18,cex=0.7)
      lines((0:2500)/100,100/(0:2500),col="blue")
      legend("topright",legend=c("won home","won draw","won away","loss home","loss draw","loss away",paste("#bets won=",sum(res==1)),paste("#bets lost=",sum(res==-1))),
             col = c(rep(c("green","red"),each=3),NA,NA),pch=c(rep(c(16,17,18),2),NA,NA),
      )
    }
    else{
      plot(odds[which(res[,1]==1),1],probs[which(res[,1]==1),1],col="green",xlim = c(0,25),pch=16,ylim = c(0,1),
           main=paste("Season=",season," and model=",model),xlab = "Odds",ylab = "Estimated Probability",cex=0.5)
      points(odds[which(res[,1]==-1),1],probs[which(res[,1]==-1),1],col="red",pch=16,cex=0.5)
      points(odds[which(res[,2]==-1),2],probs[which(res[,2]==-1),2],col="red",pch=16,cex=0.5)
      points(odds[which(res[,3]==-1),3],probs[which(res[,3]==-1),3],col="red",pch=16,cex=0.5)
      points(odds[which(res[,2]==1),2],probs[which(res[,2]==1),2],col="green",pch=16,cex=0.5)
      points(odds[which(res[,3]==1),3],probs[which(res[,3]==1),3],col="green",pch=16,cex=0.5)
      lines((0:2500)/100,100/(0:2500),col="blue")
      legend("topright",legend=c("won","lost","expectation is 1",paste("#bets won=",sum(res==1)),paste("#bets lost=",sum(res==-1))),
             col = c("green","red","blue",NA,NA),pch=c(16,16,NA,NA,NA),lty=c(NA,NA,1,NA,NA)
      )
    }
    
  }
  #returns the probabilites used by Bet365
  bet365toprobs <- function(oh,od,oa){#calculates probabilities based on odds from bet365
    alpha = 1/(1/oh+1/od+1/oa)
    p1 = alpha/oh
    p2 = alpha/od
    p3 = alpha/oa
    return(c(p1,p2,p3))#returns (P(H),P(D),P(A))
  }
  
  
  all365probs <- function(n,m,pl){
    probs = matrix(0,ncol=3,nrow = m-n+1)
    for(i in (n):m){
      probs[i-n+1,] = (bet365toprobs(pl$B365H[i],pl$B365D[i],pl$B365A[i]))
    }
    return(probs)
  }
  #plots the odds and probabilities on a log scale
  plotOddsLog <- function(odds,probs,results,season="",model=""){
    color = c("green","red")
    HUB = c("H","D","A")
    ER = odds*probs
    res = matrix(0,ncol = 3,nrow = length(results))
    for(i in 1:length(probs[,1])){
      for(j in 1:length(probs[1,])){
        if(ER[i,j]>1){
          res[i,j]=2*(results[i]==HUB[j])-1
        }
      }
    }
    plot(log(probs[which(res[,1]==1),1]),log(odds[which(res[,1]==1),1]),xlim = c(log(0.04),log(1)),ylim = c(log(1),log(25)),xaxt="n",yaxt="n",
         main=paste0("Season=",season," and model=",model),xlab = "Estimated probability",ylab = "Odds from Bet365",col="green",cex=0.7,pch=16  )
    axis(2,at=c(log(1),log(2),log(5),log(10),log(25)),c(1,2,5,10,25))
    axis(1,at=c(log(0.05),log(0.1),log(0.2),log(0.5),log(1)),c(0.05,0.1,0.2,0.5,1))
    
    points(log(probs[which(res[,1]==-1),1]),log(odds[which(res[,1]==-1),1]),col="red",pch=17,cex=0.7)
    points(log(probs[which(res[,2]==-1),2]),log(odds[which(res[,2]==-1),2]),col="red",pch=17,cex=0.7)
    points(log(probs[which(res[,3]==-1),3]),log(odds[which(res[,3]==-1),3]),col="red",pch=17,cex=0.7)
    points(log(probs[which(res[,2]==1),2]),log(odds[which(res[,2]==1),2]),col="green",pch=16,cex=0.7)
    points(log(probs[which(res[,3]==1),3]),log(odds[which(res[,3]==1),3]),col="green",pch=16,cex=0.7)
    
    legend("topright",legend=c("won","lost",paste("#bets won=",sum(res==1)),paste("#bets lost=",sum(res==-1))),
           col = c(rep(c("green","red")),NA,NA),pch=c(16,17,NA,NA)
    )
    
  }
  #assumes the correlation is not zero
  timePoisBIrhoToProbs <- function(PL,M){
    dyn.load(dynlib("BNIrho"))
    random=c("a","d")
    parameters = list(log_sd = c(0.1,0.1),
                      a=rep(0,20),d=rep(0,20),c=0.5,ch=0.5,transformed_rho=0)
    #always testing for the last half of the season
    probs = cbind(rep(0,M*10),rep(0,M*10),rep(0,M*10))
    x = PL$FTHG
    y = PL$FTAG
    HA = HomeAway(PL)
    home = as.factor(HA[,1]-1)
    away = as.factor(HA[,2]-1)
    for(i in 1:M){
      data = list(x = x[1:(190+(i-1)*10)],y = y[1:(190+(i-1)*10)],home=home[1:(190+(i-1)*10)],away=away[1:(190+(i-1)*10)],N=nlevels(home))
      obj <- MakeADFun(data=data,parameters=parameters,random = random,DLL = "BNIrho", silent = TRUE)
      res <- optim(par = obj$par, fn = obj$fn, gr = obj$gr, method = "BFGS")
      rep = sdreport(obj)
      probs[((i-1)*10+1):(i*10),] = poisToProbs(home[(1+190+(i-1)*10):(190+(i)*10)],away[(1+190+(i-1)*10):(190+(i)*10)],rep)
    }
    return(probs)
  }
  #assumes that the home fields advantage is not constant over the whole time period. 
  fullToProbsTauMLECH <- function(PL,start,slutt,n1,M1,NS){
    
    #start er frste runde som skal ha sannsynligheter 
    #slutt er site runde som skal ha sannsynligheter 
    #start=20,slutt=38. gir sannsynligheter for siste halvdel av sesongen
    #n1 er antall kamper som ikke skal vre med for prediksjonen. n1=0 gir hele datasettet
    #PL[(n1+1):M1,] er datasettet som skal brukes til frste prediksjon
    #for  bruke en sesong og prediktere andre halvdel s skal n1=0 og m1=190
    #NS=2 hvis det ikke er med suportere, det vil si for siste sesong
    sos1 = rep(1,12*380)
    sos1[1:(10*380+288)]=0
    probs = matrix(0,ncol = 3,nrow = (slutt-start+1)*10)
    #i=1
    startTime = Sys.time()
    temp = Sys.time()
    k=slutt-start+1
    for(i in 1:k){
      print(i)
      print( Sys.time()-temp)
      temp = Sys.time()
      EPL = PL[(n1+1):(M1+10*(i-1)),]
      EPL2 = PL[(M1+10*(i-1)+1):(M1+10*(i)),]
      teams = levels(as.factor(PL$HomeTeam[1:(M1+10*(k))]))
      nt = length(teams)#the number of different teams in the dataset
      n = length(c(EPL$Date[which(EPL$HomeTeam==teams[1])],EPL$Date[which(EPL$AwayTeam==teams[1])]))#the total number of matches played by each team that has been in the top every season
      for(j in 2:nt){
        n = max(n, length(c(EPL$Date[which(EPL$HomeTeam==teams[j])],EPL$Date[which(EPL$AwayTeam==teams[j])])))
      }
      HA = HomeAwayFull(EPL,teams)
      home = as.factor(HA[,1]-1)
      away = as.factor(HA[,2]-1)
      x = EPL$FTHG
      y = EPL$FTAG
      
      data = list(x=x,y=y,home=home,away=away,nt=nt)
      parameters = list(a0=rep(0,nt),d0=rep(0,nt),log_sd=c(-1,-1),c=0.5,ch=0.5)
      
      random = c("a0","d0")
      
      obj <- MakeADFun(data=data,parameters=parameters,random = random,DLL = "FullModelConst", silent = TRUE)
      res <- optim(par = obj$par, fn = obj$fn, gr = obj$gr, method = "BFGS")
      
      repC = sdreport(obj)
      #print(repC)
      #sn = as.factor(rep(0:1,each=4088))
      a0 = repC$par.random[1:nt]
      d0 = repC$par.random[(nt+1):(2*nt)]
      log_sd = repC$par.fixed[1:2]
      c = repC$par.fixed[3]
      ch = repC$par.fixed[4]
      a = matrix(0,nrow = nt,ncol=n-1)
      d = matrix(0,nrow = nt,ncol=n-1)
      for(j in 1:nt){
        a[j,] = rep(a0[j],n-1)
        d[j,] = rep(d0[j],n-1)
      }
      
      HA = HomeAwayFull(EPL,teams)
      m = matchNumber(EPL,teams)
      sos = sos1[1:length(EPL$HomeTeam)]
      data = list(x = EPL$FTHG,y=EPL$FTAG,home = as.factor(HA[,1]-1),away=as.factor(HA[,2]-1), DT = sqrt(makeDTfull(EPL,teams,n)),
                  mh = as.factor(m[,1]),ma = as.factor(m[,2]),N=n-1,nt=nt,sos=sos)
      
      parameters = list(a0=a0,d0= d0,a = a,d = d,log_sd=log_sd,
                        c=rep(c,2),ch=rep(ch,2),tau=55)
      
      random = c("a0","d0","a","d")
      
      obj <- MakeADFun(data=data,parameters=parameters,random = random,DLL = "FullModelCh", silent = TRUE)
      res <- optim(par = obj$par, fn = obj$fn, gr = obj$gr, method = "BFGS")
      rep = sdreport(obj)
      
      a = matrix(rep$par.random[(2*nt+1):(2*nt+nt*(n-1))],ncol=n-1,nrow = nt)
      a0 = rep$par.random[1:nt]
      d0 = rep$par.random[(nt+1):(2*nt)]
      d = matrix(rep$par.random[(2*nt+nt*(n-1)+1):(2*nt+2*nt*(n-1))],ncol=n-1,nrow = nt)
      c = rep$par.fixed[3:4]
      ch = rep$par.fixed[5:6]
      
      m1 = matchNumber(rbind(EPL,EPL2),teams)
      mh1 = pmin(m1[(M1+10*(i-1)+1):(M1+10*(i)),1]-1,n-1)
      ma1 = pmin(m1[(M1+10*(i-1)+1):(M1+10*(i)),2]-1,n-1)
      HA1 = HomeAwayFull(EPL2,teams)
      home1 = HA1[,1]
      away1 = HA1[,2]
      for(j in 1:10){
        lamh = (a[home1[j],mh1[j]]-d[away1[j],ma1[j]]+c[NS]+ch[NS])
        lama = (a[away1[j],ma1[j]]-d[home1[j],mh1[j]]+c[NS])
        
        probs[(i-1)*10+j,]=cbind(1-pskellam(0,exp(lamh),exp(lama)),dskellam(0,exp(lamh),exp(lama)),pskellam(-1,exp(lamh),exp(lama)))
      }
      
    }
    #return(cbind(a0,d0,a,d))
    print(Sys.time()-startTime)
    return(probs)
  }
  
  
  fullToProbsTauCH <- function(PL,start,slutt,n1,M1,tau,NS=1){
    #start er frste runde som skal ha sannsynligheter 
    #slutt er site runde som skal ha sannsynligheter 
    #start=20,slutt=38. gir sannsynligheter for siste halvdel av sesongen
    #n1 er antall kamper som ikke skal vre med for prediksjonen. n1=0 gir hele datasettet
    #PL[(n1+1):M1,] er datasettet som skal brukes til frste prediksjon
    #for  bruke en sesong og prediktere andre halvdel s skal n1=0 og m1=190
    #NS=2 hvis det ikke er med suportere, det vil si for siste sesong
    sos1 = rep(1,12*380)
    sos1[1:(10*380+288)]=0
    probs = matrix(0,ncol = 3,nrow = (slutt-start+1)*10)
    #i=8
    startTime = Sys.time()
    temp = Sys.time()
    k=slutt-start+1
    for(i in 1:k){
      print(i)
      print( Sys.time()-temp)
      temp = Sys.time()
      EPL = PL[(n1+1):(M1+10*(i-1)),]
      EPL2 = PL[(M1+10*(i-1)+1):(M1+10*(i)),]
      teams = levels(as.factor(PL$HomeTeam[1:(M1+10*(k))]))
      nt = length(teams)#the number of different teams in the dataset
      n = length(c(EPL$Date[which(EPL$HomeTeam==teams[1])],EPL$Date[which(EPL$AwayTeam==teams[1])]))#the total number of matches played by each team that has been in the top every season
      for(j in 2:nt){
        n = max(n, length(c(EPL$Date[which(EPL$HomeTeam==teams[j])],EPL$Date[which(EPL$AwayTeam==teams[j])])))
      }
      HA = HomeAwayFull(EPL,teams)
      home = as.factor(HA[,1]-1)
      away = as.factor(HA[,2]-1)
      x = EPL$FTHG
      y = EPL$FTAG
      
      data = list(x=x,y=y,home=home,away=away,nt=nt)
      parameters = list(a0=rep(0,nt),d0=rep(0,nt),log_sd=c(-1,-1),c=0.5,ch=0.5)
      
      random = c("a0","d0")
      
      obj <- MakeADFun(data=data,parameters=parameters,random = random,DLL = "FullModelConst", silent = TRUE)
      res <- optim(par = obj$par, fn = obj$fn, gr = obj$gr, method = "BFGS")
      
      repC = sdreport(obj)
      #print(repC)
      #sn = as.factor(rep(0:1,each=4088))
      a0 = repC$par.random[1:nt]
      d0 = repC$par.random[(nt+1):(2*nt)]
      log_sd = repC$par.fixed[1:2]
      c = repC$par.fixed[3]
      ch = repC$par.fixed[4]
      a = matrix(0,nrow = nt,ncol=n-1)
      d = matrix(0,nrow = nt,ncol=n-1)
      for(j in 1:nt){
        a[j,] = rep(a0[j],n-1)
        d[j,] = rep(d0[j],n-1)
      }
      
      HA = HomeAwayFull(EPL,teams)
      m = matchNumber(EPL,teams)
      sos = sos1[1:length(EPL$HomeTeam)]
      data = list(x = EPL$FTHG,y=EPL$FTAG,home = as.factor(HA[,1]-1),away=as.factor(HA[,2]-1),DT = sqrt(makeDTfull(EPL,teams,n)),
                  mh = as.factor(m[,1]),ma = as.factor(m[,2]),N=n-1,nt=nt,tau=tau,sos=sos)
      
      parameters = list(a0=a0,d0= d0,a = a,d = d,log_sd=log_sd,
                        c=rep(c,2),ch=rep(ch,2))
      
      random = c("a0","d0","a","d")
      
      obj <- MakeADFun(data=data,parameters=parameters,random = random,DLL = "FullModelTauCh", silent = TRUE)
      res <- optim(par = obj$par, fn = obj$fn, gr = obj$gr, method = "BFGS")
      rep = sdreport(obj)
      
      a = matrix(rep$par.random[(2*nt+1):(2*nt+nt*(n-1))],ncol=n-1,nrow = nt)
      a0 = rep$par.random[1:nt]
      d0 = rep$par.random[(nt+1):(2*nt)]
      d = matrix(rep$par.random[(2*nt+nt*(n-1)+1):(2*nt+2*nt*(n-1))],ncol=n-1,nrow = nt)
      c = rep$par.fixed[3:4]
      ch = rep$par.fixed[5:6]
      
      m1 = matchNumber(rbind(EPL,EPL2),teams)
      mh1 = pmin(m1[(M1+10*(i-1)+1):(M1+10*(i)),1]-1,n-1)
      ma1 = pmin(m1[(M1+10*(i-1)+1):(M1+10*(i)),2]-1,n-1)
      HA1 = HomeAwayFull(EPL2,teams)
      home1 = HA1[,1]
      away1 = HA1[,2]
      for(j in 1:10){
        lamh = (a[home1[j],mh1[j]]-d[away1[j],ma1[j]]+c[NS]+ch[NS])
        lama = (a[away1[j],ma1[j]]-d[home1[j],mh1[j]]+c[NS])
        
        probs[(i-1)*10+j,]=cbind(1-pskellam(0,exp(lamh),exp(lama)),dskellam(0,exp(lamh),exp(lama)),pskellam(-1,exp(lamh),exp(lama)))
      }
      
    }
    #return(cbind(a0,d0,a,d))
    print(Sys.time()-startTime)
    return(probs)
  }
  
}
