#This file was used to generate the results found in the Home field advantage section of the project. 
dyn.load(dynlib("binormalintercept"))
dyn.load(dynlib("multinomRI"))
seasons = c("0910","1011","1112","1213","1314","1415","1516","1617","1718","1819","1920","2021")
sesonger = c("09/10","10/11","11/12","12/13","13/14","14/15","15/16","16/17","17/18","18/19","19/20","20/21")
c = rep(0,length(seasons))#a vector that will have the estimed value for c for all seasons
ce = rep(0,length(seasons))#a vector with the standard devaition for c 
ch = rep(0,length(seasons))#a vector with the estimates for ch
che = rep(0,length(seasons))#standard deviation for ch
cche = rep(0,length(seasons))#the covariance between c and ch
log_sdA = rep(0,length(seasons))
log_sdD = rep(0,length(seasons))
ph = rep(0,length(seasons))#estimated probability for home victory
pa = rep(0,length(seasons))#estimated probability for away victory

for(i in 1:length(seasons)){
  PL = eval(as.symbol(paste0("pl_",seasons[i])))
  x = PL$FTHG
  y = PL$FTAG
  HA = HomeAway(PL)
  home = as.factor(HA[,1]-1)
  away = as.factor(HA[,2]-1)
  N = nlevels(home)
  data = list(x=x,y=y,home=home,away=away,N=N)
  parameters = list(c=0.1,ch=0.1,a=rep(0,N),d = rep(0,N),log_sd=c(-1,-1))
  random = c("a","d")
  
  obj <- MakeADFun(data=data,parameters=parameters,random = random,DLL = "binormalintercept", silent = TRUE)
  resBI <- optim(par = obj$par, fn = obj$fn, gr = obj$gr, method = "BFGS")
  repBI = sdreport(obj)
  c[i] = repBI$par.fixed[3]
  ce[i] = sqrt(repBI$cov.fixed[3,3])
  ch[i] = repBI$par.fixed[4]
  che[i] = sqrt(repBI$cov.fixed[4,4])
  cche[i] = repBI$cov.fixed[3,4]
  log_sdA[i] = repBI$par.fixed[1]
  log_sdD[i] = repBI$par.fixed[2]
  
  X = cbind(as.numeric(PL$FTR=="H"),as.numeric(PL$FTR=="D"),as.numeric(PL$FTR=="A"))
  data = list(X=X,home=home,away=away)
  parameters = list(theta = 0.1,h=0,lam=rep(0,N),log_sigma=-1)
  random = c("lam")
  obj <- MakeADFun(data=data,parameters=parameters,random = random,DLL = "multinomRI", silent = TRUE)
  resRI <- optim(par = obj$par, fn = obj$fn, gr = obj$gr, method = "BFGS")
  repRI = sdreport(obj)
  
  ph[i] = plogis(-repRI$par.fixed[1]+repRI$par.fixed[2])
  pa[i] = 1- plogis(repRI$par.fixed[1]+repRI$par.fixed[2])
}

#plotting the home field advantage in percent.
plot((exp(ch)-1)*100,type = "b",col="blue",ylim = c(-10,80),xlab = "Season",xaxt="n",ylab = "Percent increase in number og goals scored at home",
     main="The home field advantage in percent")
axis(1,at=1:12,sesonger,las=2)
lines((exp(ch+1.96*che)-1)*100,type = "b",col="red")
lines((exp(ch-1.96*che)-1)*100,type = "b",col="red")
abline(h=0)
legend("topright",legend = c("Estimated percent increase","95% confidence interval"),col=c("blue","red"),lty = 1)

#plotting expected goals scored by home and away team
var = (exp(log_sdA))^2+(exp(log_sdD))^2
plot(exp(c+ch+var/2),type = "b",ylim = c(0.7,2.3),col="blue",lwd=2,main="Expected number of goals with unknown teams",xaxt="n",xlab="seasons",
     ylab = "Expected number of goals scored per match")
axis(1,at=1:12,sesonger,las=2)
lines(exp(c+ch+var/2+1.96*sqrt(ce^2+che^2+2*cche)),type = "b",col="lightblue")
lines(exp(c+ch+var/2-1.96*sqrt(ce^2+che^2+2*cche)),type = "b",col="lightblue")
lines(exp(c+var/2),type = "b",col="red",lwd=2)
lines(exp(c+var/2+1.96*ce),type = "b",col="pink")
lines(exp(c+var/2-1.96*ce),type = "b",col="pink")
legend("topright",legend = c("Expected number of goals scored home","Expected number of goals scored away","95% confidence interval home",
                             "95% confidence interval away"),lty = 1,lwd=c(2,2,1,1),col=c("blue","red","lightblue","pink"))


#plotting the probability for home win away win and draw for the different seaosons

plot(ph,type = "b",ylim = c(0,0.6),col="blue",xaxt="n",xlab = "Seasons",ylab = "Probability",
     main = "Estimated probability for different outcomes.")
axis(1,at=1:12,sesonger,las=2)
lines(pa,type = "b",col="red")
lines(1-pa-ph,type = "b",col="purple")
legend("bottomright",c("Probability of home victory","Probability of draw","Probability of away victory"),lty=1,col=c("blue","purple","red"))

plot(ph/pa,type = "b",xaxt="n",xlab="Season",ylab="P(H)/P(A)",main = "Ratio of home win to away win",col="blue")
axis(1,at=1:12,sesonger,las=2,col="blue")
abline(h=1,col="red")











