#include <TMB.hpp>
/*This file creates the negative log-likelihood for the data intensive function with the same parameter alpha for all 
 * the teams. It will only be used to test the hypotesis that alpha is the same for all the teams
 */

template<class Type>
Type objective_function<Type>::operator() ()
{
  //all the vectors are of length equal to the number of matches.
  DATA_VECTOR(x); // home goals
  DATA_VECTOR(y); // away goals
  DATA_VECTOR(Th); // shots on target home team
  DATA_VECTOR(Ta);  // shots on target away team
  DATA_VECTOR(Sh); // shots home team
  DATA_VECTOR(Sa);  // shots away team
  DATA_FACTOR(home); // what teams playes at home
  DATA_FACTOR(away); // what teams playes away 
  
  PARAMETER(a); //pnorm(a) is the probability that a shot will hit target 
  PARAMETER_VECTOR(b); // 1-pnorm(b_i) is the probability that team i will manage to defend away a chanse. so it does not become a shot
  PARAMETER_VECTOR(g);//1-pnomr(g_i) is the probability that team i will save a shot on target
  PARAMETER_VECTOR(log_lambda);  // lambda_i is the expected number of chanses team i will get in a game
  PARAMETER(log_h);// h*lambda_i is the expected number of chanses team i will get at home
  
  
  Type nll = 0; //the negative log likelihood
  vector<Type> lambda = exp(log_lambda);
  Type lamH; //expected number of shots for home team
  Type lamA;//expected numebr of shots for away team
  Type pH;//probability that a shot hits target for the home team
  Type pA;//probability that a shot hits target for the away team
  Type p1H;//probability that a shot on target goes in for the home team
  Type p1A;//probability that a shot on target goes in for the away team
  
  //will iterate for each match
  //and every observation will be included in the negative log likeliood as 
  //described in the project
  for(int i = 0;i < x.size();i++){
    lamH = (exp(log_h)*lambda[home[i]])*pnorm(b[away[i]]);
    nll -= dpois(Sh[i],lamH,true);
    pH = pnorm(a);
    nll -= dbinom(Th[i],Sh[i],pH,true);
    p1H = pnorm(g[away[i]]);
    nll -= dbinom(x[i],Th[i],p1H,true);
    
    lamA = (lambda[away[i]])*pnorm(b[home[i]]);
    nll -= dpois(Sa[i],lamA,true);
    pA = pnorm(a);
    nll -= dbinom(Ta[i],Sa[i],pA,true);
    p1A = pnorm(g[home[i]]);
    nll -= dbinom(y[i],Ta[i],p1A,true);
    
  }
  
  return nll;
}