#Everything that has to do with the Poisson GLMM described in the project
compile("binormalintercept.cpp")
dyn.load(dynlib("binormalintercept"))

season = 2021# assigning what seasons you would like to study
sesong = as.character(season)
PL = eval(as.symbol(paste0("pl_",season))) # assigning what seasons you would like to study
M = 19#the number of rounds after round 19. This was mainly a nesesaty when the current season(20/21) was not finished yet

teams = levels(as.factor(PL$HomeTeam))
x = PL$FTHG
y = PL$FTAG
HA = HomeAway(PL)
home = as.factor(HA[,1]-1)
away = as.factor(HA[,2]-1)
N = nlevels(home)
data = list(x=x,y=y,home=home,away=away,N=N)#the data to be included when using TMB, they corespoond to the c++ file "binormalintercept"
parameters = list(c=0.1,ch=0.1,a=rep(0,N),d = rep(0,N),log_sd=c(-1,-1))#the parameters to be estimated by TMB
random = c("a","d")#the parameters that should be considered random, they will be integrated out in the likelihood using the laplace approximation

obj <- MakeADFun(data=data,parameters=parameters,random = random,DLL = "binormalintercept", silent = TRUE)#creating the object using TMB
resBI <- optim(par = obj$par, fn = obj$fn, gr = obj$gr, method = "BFGS")#optimice the object using standard optimazation in R

repBI = sdreport(obj)#the parameters 
aBI = repBI$par.random[1:N]#the attacking strength, par.random gives a vector with all the random effects
dBI = repBI$par.random[(N+1):(2*N)]
aBIe = repBI$diag.cov.random[1:N]#the variance for each of the random effects
dBIe = repBI$diag.cov.random[(N+1):(2*N)]

#plotting the attacking strength for this model on the chosen season
plot(exp(aBI),type = "b",xaxt="n",xlab = NA,col="blue",ylim = c(0.2,2.5),main = paste0("Poisson GLMM on season = ",substring(sesong,1,2),"/",substring(sesong,3,4)),
     ylab = "Multiplier to the expacted number of goals scored")
lines(exp(aBI+1.96*sqrt(aBIe)),type = "b",col="red")
lines(exp(aBI-1.96*sqrt(aBIe)),type = "b",col="red")
axis(1,at=1:20,teams,las=2,cex.axis=0.7,col="blue")
legend("topleft",c("Estimated multiplier","95% confidence interval"),col = c("blue","red"),lty = 1)

#plotting the defending strength for this model on the chosen season
plot(exp(-dBI),type = "b",xaxt="n",xlab = NA,col="blue",ylim = c(0.5,2),main = paste0("Poisson GLMM on season = ",substring(sesong,1,2),"/",substring(sesong,3,4)),
     ylab = "Multiplier to the expacted number of goals conceded")
lines(exp(-dBI-1.96*sqrt(dBIe)),type = "b",col="red")
lines(exp(-dBI+1.96*sqrt(dBIe)),type = "b",col="red")
axis(1,at=1:20,teams,las=2,cex.axis=0.7,col="blue")
legend("topleft",c("Estimated multiplier","95% confidence interval"),col = c("blue","red"),lty = 1)


#testing for covariance between attack and defence

parameters = list(c=0.1,ch=0.1,a=rep(0,N),d = rep(0,N),log_sd=c(-1,-1),transformed_rho=0)
obj <- MakeADFun(data=data,parameters=parameters,random = random,DLL = "BNIrho", silent = TRUE)#creating the object using TMB
resRho <- optim(par = obj$par, fn = obj$fn, gr = obj$gr, method = "BFGS")#optimize the object using standard optimazation in R
repRho = sdreport(obj)
2*plogis(repRho$par.fixed[5])-1#very high value for all the seasons
d = 2*(resBI$value-resRho$value)
1-pchisq(d,1)#the p-value whith null-hypotesis of rho=0.
################################################################################################################
probs = timePoisBIToProbs(PL,M)#the probability for the first M rounds after round 19, in the season indicated by PL
odds= cbind(PL$B365H,PL$B365D,PL$B365A)[191:(190+M*10),]#the odds corresponding to  the probabilities
results = PL$FTR[191:(190+M*10)]#the results corresponding to the probabiliites and odds.


#the results using fixed bets
coin = fixedBet(probs,odds,1,results)
plot(coin,type = "l",col="red",ylim=c(-5,15),lwd=2,main = paste0("Fixed bet. Season = ",substring(sesong,1,2),"/",substring(sesong,3,4)),xlab = "#rounds after round 19",ylab = "Money earned" )
for(i in 1:20){
  coinA = fixedBet(probs,odds,1,randomResults(probs))
  lines(coinA,col=greys[i])
}
abline(h=0,col="blue")
legend("topleft",legend = c("The betting results","line at zero","different realizations"),
       lwd=c(2,1,1),lty=1,col=c("red","blue",greys[1]))

coin[M+1]/M*100#the return on investment in percent

#the results using fixed return 

coin = FixedReturn(probs,odds,1,results)
plot(coin,type = "l",ylim=c(-5,10),col="red",lwd=3,main = paste0("Fixed return. Season = ",substring(sesong,1,2),"/",substring(sesong,3,4)),
     xlab = "Rounds of betting",ylab = "Money earned")
for(i in 1:20){
  coinA = fixedBet(probs,odds,1,randomResults(probs))
  lines(coinA,col=greys[i])
}
abline(h=0,col="blue")
legend("topleft",legend = c("The betting results","line at zero","different realizations"),
       lwd=c(2,1,1),lty=1,col=c("red","blue",greys[1]))
coin[M+1]/M*100

#the results  using rue and salvesen betting strategy 
coin = RueVariance(probs,odds,1,results)
plot(0:19,coin,type = "l",col="red",lwd=3,ylim = c(-4,10),ylab = "Money earned",xlab="Rounds of betting",main=paste0("Rue & Salvesen,model: Poisson GLMM, season = ",substring(sesong,1,2),"/",substring(sesong,3,4)))
for(i in 1:20){
  coinA = RueVariance(probs,odds,1,randomResults(probs))
  lines(0:M,coinA,col=greys[i])
}
abline(h=0,lty=3,col="blue")
legend("topleft",legend = c("Money earned","Zero","Different realizations"),lty = c(1,3,1),lwd=c(3,1,1),
       col = c("red","blue","grey"))
coin[M+1]/M*100

(testKelly(probs,odds,1000,results,0.2)[20]-1000)/1000*100

################################################################################################################################
#the odds and estimated probability is plotted together in here


plotOddsLog(odds,probs,results,season = paste0(substring(sesong,1,2),"/",substring(sesong,3,4)),model = "Poisson GLMM")





##################################################################################################################################
#need to run the file named "check_overdispersion.R" first.
library(glmmTMB)
goals = c(EPL$FTHG,EPL$FTAG)
attack = c(EPL$HomeTeam,EPL$AwayTeam)
defence = c(EPL$AwayTeam,EPL$HomeTeam)
h = rep(c(1,0),each = length(EPL$HomeTeam))
data = data.frame(goals=goals,attack=attack,defence=defence,h=h)
test = glmmTMB(goals~(1|attack)-(1|defence)+h,data=data,family = poisson(link = "log"))

check_overdispersion.glmmTMB(test)#this test for over-dispertion, there is no indication of over-dispertion.























